# vim:fileencoding=utf-8
"""basewindows.py

"""
#   kuechenstation - a curses client for xmms2
#   
#   Copyright (C) 2008  Jan Eike von Seggern <eikevons@yahoo.de>
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import curses
import logging
import ConfigParser

from eventframework import EventRecipient
from helpers import ms2string, needsitems

class SizeError(Exception):
    pass


class MultiWindow(EventRecipient):
    """Base window for a client window in the multi window region."""
    def __init__(self, name, parent, dist, window, mainapp):
        super(MultiWindow, self).__init__(parent, dist)
        self.name = name
        self.__win = window
        self.__mainapp = mainapp
        self.encoding = self.__mainapp.getOption('general', 'encoding')

        self.colorattr = self.__mainapp.getColorAttr(self.name)
        logging.info("%s using color attr #%i"%(self.name, self.colorattr))

    def message(self, m):
        self.__mainapp.setVolatileText(m)

    def extraMessage(self, m):
        self.__mainapp.setExtraText(m)

    def refresh(self):
        if not self.active:
            raise RuntimeError("Refreshing non active MultiWindow"
                               " '%s'!"%(self))
        self.window.refresh()

    def activate(self):
        super(MultiWindow, self).activate()
        self.window.clear()
        MultiWindow.refresh(self)
        self.message("activated %s"%(self.__class__.__name__))

    def setWindow(self, newwin):
        self.__win = newwin

    def addLine(self, y, *elements):
        """Add a line at y to the screen.

        It's possible that not all items in elements are printed.

        elements -- the elements to put on the line. Can be a string or a
            list of (string, curses attribute, [left|right]) tuples.
        """
        logging.debug("addLine(%s, %s)" %(y, elements))
        x = 0
        for el in elements:
            if type(el) in (unicode, str):
                x += self.addStr(y, x, s=el)
            elif len(el) == 2:
                x += self.addStr(y, x, s=el[0], attr=el[1])
            else:
                x += self.addStr(y, x, s=el[0], attr=el[1], strip=el[2])
        if x < self.width:
            self.addStr(y, x, " " * (self.width - x))

    def addStr(self, y, x, s, attr=None, strip="right"):
        """Adds the given string with the given attributes to the screen.
        Returns the length of the *encoded* string.

        s -- the string, can be str or unicode type
        strip -- decides where to put "..." if s is to long for the desired
            space, can be "right" or "left" or "none".
        """
        # logging.debug("addStr(%s, %s, '%s', %s, %s)" % (y, x, s, attr, strip))
        if type(s) is unicode:
            s = s.encode(self.encoding, "replace")
        if len(s) > (self.width - x):
            if strip == "left":
                s = "..." + s[-self.width + x + 3:]
            elif strip == "right":
                s = s[:self.width - x - 3] + "..."
            else:
                s = s[:self.width - x]

        if attr is None:
            attr = self.colorattr

        self.__win.addstr(y, x, s, attr)
        return len(s)

    mainapp = property(lambda self: self.__mainapp)
    xmms = property(lambda self: self.__mainapp.xmms)
    window = property(lambda self: self.__win, setWindow)
    # TODO: muss -1 gemacht werden? => nein?
    height = property(lambda self: self.__win.getmaxyx()[0])
    # TODO: muss -1 gemacht werden? => nein
    width = property(lambda self: self.__win.getmaxyx()[1])


class ListItem(object):
    def __init__(self, data, encoding):
        self.data = data
        self.selected = False
        self.encoding = encoding

    def select(self, b=True):
        self.selected = b

    def toggleSelect(self):
        self.selected = not self.selected

    def getString(self, width):
        if self.selected:
            mark = u"*"
        else:
            mark = u" "

        l = u"%s %s"%(mark, self.data)

        l = l.encode(self.encoding, 'replace')

        # do all the size stuff with encoded ascii strings
        if len(l) < width:
            l = l.ljust(width)
        elif len(l) > width:
            l = l[:width]
        return l


class ListWindow(MultiWindow):
    items = None
    __offset = 0
    # the list index of the item currently selected by the cursor
    __cursel = 0
    heightreduce = 0
    def __init__(self, *multiwinargs):
        super(ListWindow, self).__init__(*multiwinargs)

        self.__init_register_events()
        self.cursorattr = self.mainapp.getAttrOption(self.name, "cursor attr")

    def __init_register_events(self):
        rE = self.registerEvent

        rE(self.evCursorUp, curses.KEY_UP, 'k')
        rE(self.evCursorDown, curses.KEY_DOWN, 'j')
        rE(self.evPageUp, curses.KEY_PPAGE)
        rE(self.evPageDown, curses.KEY_NPAGE)
        rE(self.evPosHome, curses.KEY_HOME)
        rE(self.evPosEnd, curses.KEY_END)

        rE(self.evSelect, ' ')

    def setOffset(self, new):
        logging.debug("ListWindow.setOffset:...")
        if new > len(self.items) - self.height + self.heightreduce:
            new = len(self.items) - self.height + self.heightreduce
        if new < 0:
            new = 0
        self.__offset = new
    offset = property(lambda self: self.__offset, setOffset)

    def setCursel(self, new):
        if new < 0:
            new = 0
        elif new >= len(self.items):
            new = len(self.items) - 1
        self.__cursel = new
        self.extraMessage("%i/%i"%(new + 1, len(self.items)))
    cursel = property(lambda self: self.__cursel, setCursel)

    def repairIndices(self):
        # checks are done by setOffset and setCursel
        self.cursel = self.cursel
        self.offset = self.offset

        if self.offset > self.cursel:
            logging.debug("resetting offset: %i"%(self.cursel))
            self.offset = self.cursel
            logging.debug("resulting offset: %i"%(self.offset))

    def addAttr(self, index, attr):
        if index == self.cursel:
            attr |= self.cursorattr
        attr |= self.colorattr
        return attr

    def displayItem(self, index, y=None, attr=0):
        # logging.debug("displayItem: y=%i, height=%i,"
                      # " index=%i, offset=%i, len=%i"%(y, self.height, index,
                                              # self.offset, len(self.items)))
        if y is None:
            y = index - self.offset
        if y >= self.height or y < 0:
            logging.error("y out of range")
            raise SizeError("item %i is not in display area:"
                            " y=%i!"%(index, y))

        if index >= len(self.items):
            logging.error("list index out of range")
            raise SizeError("list index %i out of range, y=%i!"%(index, y))

        line = self.items[index].getString(self.width)

        attr = self.addAttr(index, attr)

        try:
            self.addLine(y, (line, attr))
            # self.window.addstr(y, 0, line, attr)
        except curses.error:
            logging.error("error: displayItem: y=%i, height=%i,"
                      " index=%i, offset=%i, len=%i"%(y, self.height, index,
                                              self.offset, len(self.items)))
            pass

    def refresh(self):
        if not self.active:
            return

        self.repairIndices()
        self.window.clear()

        for y in xrange(self.heightreduce, self.height):
            try:
                index = y + self.offset
                if index < len(self.items):
                    self.displayItem(index, y)
                else:
                    self.addLine(y)
            except curses.error, e:
                logging.error("Catched curses.error: %s,"
                        " y = %i, index = %i, height = %i"
                        % (e, y, index, self.height))


        # call __win.refresh()
        super(ListWindow, self).refresh()

    # events
    @needsitems
    def evSelect(self, event):
        self.items[self.cursel].toggleSelect()
        self.displayItem(self.cursel)
        super(ListWindow, self).refresh()
        self.message("selected item %i"%(self.cursel))

    # cursor movement
    @needsitems
    def evCursorUp(self, event):
        if self.cursel == 0:
            self.message("reached top of list")
            return

        self.cursel -= 1
        if self.cursel < self.offset:
            self.offset -= 1
            self.refresh()
        else:
            self.displayItem(self.cursel)
            self.displayItem(self.cursel + 1)
            # call __win.refresh()
            super(ListWindow, self).refresh()
        self.message("cursor up")

    @needsitems
    def evCursorDown(self, event):
        # handle end of list and empty list
        if self.cursel >= len(self.items) - 1:
            self.message("reached end of list")
            return

        self.cursel += 1
        if self.cursel >= self.offset + self.height:
            self.offset += 1
            self.refresh()
        else:
            self.displayItem(self.cursel)
            self.displayItem(self.cursel - 1)

            # call __win.refresh()
            super(ListWindow, self).refresh()
        self.message("cursor down")

    @needsitems
    def evPageUp(self, event):
        if self.cursel == 0:
            self.message("reached top of list")
            return

        self.cursel -= self.height
        self.offset -= self.height
        self.refresh()
        self.message("page up")

    @needsitems
    def evPageDown(self, event):
        if self.cursel == len(self.items) - 1:
            self.message("reached end of list")
            return

        self.cursel += self.height
        self.offset += self.height
        self.refresh()
        self.message("page down")

    @needsitems
    def evPosHome(self, event):
        self.cursel = 0
        self.offset = 0
        self.refresh()
        self.message("cursor home")

    @needsitems
    def evPosEnd(self, event):
        if self.cursel == len(self.items) - 1:
            self.message("reached end of list")
            return
        self.cursel = len(self.items) - 1
        self.offset = len(self.items)
        self.refresh()
        self.message("cursor end")


class TreeError(Exception):
    pass


# item +-> children(items)
#      |
#      +-> next/prev: property returning the next item for display
#      |
#      +-> data (a string or something): the data represented by this item,
#      |   e.g. the artist, always lower case, used for sorting
#      |
#      +-> getString: method to return a string for this entry
#      |
#     +-> expanded (property)
class TreeItem(object):
    # indentation for display
    indent = ''
    def __init__(self, data, encoding, index, parent=None, children=None, next=None, prev=None):
        # constant properties
        self.__data = data
        self.__parent = parent
        self.__index = index
        # init properties
        self.__children = []
        self.__next = None
        self.__prev = None
        self.__expanded = False

        self.encoding = encoding

        #set properties using set... methods
        if next:
            self.next = next
        if prev:
            self.prev = prev
        if children:
            self.addChildren(*children)

        # if type(self.__data) in (str, unicode):
            # self.__cmpdata = self.__data.lower()
        # else:
            # self.__cmpdata = self.__data['tracknr']
            # if self.__cmpdata is None:
                # self.__cmpdata = self.__data['id']

    data = property(lambda self: self.__data)
    index = property(lambda self: self.__index)

    def getParent(self):
        if self.__parent:
            return self.__parent
        raise TreeError("%s: no parent found!"%(self))
    parent = property(getParent)

    def addChildren(self, *newchildren):
        self.__children.extend(newchildren)
    children = property(lambda self: self.__children)

    def getLastOpenEntry(self):
        if self.__children and self.__expanded:
            return self.__children[-1].getLastOpenEntry()
        return self

    def setExpanded(self, new):
        if self.__children:
            self.__expanded = new
        else:
            # silently unset expanded value if we have no children and
            # therefore have nothing to expand
            self.__expanded = False
    expanded = property(lambda self: self.__expanded, setExpanded)

    def toggleExpanded(self):
        self.expanded = not self.expanded

    def getPrev(self, childok=True):
        if self.__prev:
            if childok:
                return self.__prev.getLastOpenEntry()
            else:
                return self.__prev
        elif self.__parent:
            return self.__parent
        else:
            raise TreeError("%s: no prev element found!"%(self))

    def setPrev(self, new):
        if type(self) != type(new):
            raise TreeError("prev item must be of same type:"
                            " %s != %s!"%(type(self), type(new)))
        self.__prev = new

    prev = property(getPrev, setPrev)

    def setNext(self, new):
        if type(self) != type(new):
            raise TreeError("next item must be of same type:"
                            " %s != %s!"%(type(self), type(new)))
        self.__next = new

    def getNext(self, childok=True):
        if self.expanded and childok and self.__children:
            return self.__children[0]
        elif self.__next is not None:
            return self.__next
        elif self.__parent is not None:
            return self.__parent.getNext(childok=False)
        else:
            raise TreeError("%s: no next element found!"%(self))

    next = property(getNext, setNext)

    def getString(self, width=70):
        mark = u"-"
        if not self.expanded:
            mark = u"+"
        l = u"%s %s %s"%(self.indent, mark, self.data)
        l = l.encode(self.encoding, 'replace')
        if len(l) < width:
            l = l.ljust(width)
        elif len(l) > width:
            l = l[:width]
        return l


# TODO: rewrite up/down or refresh because now the whole list is redrawn and
# that's causing flickering
class TreeWindow(ListWindow):
    # heightreduce = 0
    def __init__(self, *multiwinargs):
        super(TreeWindow, self).__init__(*multiwinargs)
        self.items = []

        # the current selected line in screen coords
        # used to check if we leave the screen with cursor movement
        self.cury = self.heightreduce

        self.__init_register_events()

    def __init_register_events(self):
        rE = self.registerEvent

        # unset Page-Up, -Down and select events
        rE(None, curses.KEY_PPAGE)
        rE(None, curses.KEY_NPAGE)
        rE(None, ' ')

        rE(self.evCursorUp, curses.KEY_UP, 'k')
        rE(self.evCursorDown, curses.KEY_DOWN, 'j')
        rE(self.evPosTop, curses.KEY_HOME)
        rE(self.evPosEnd, curses.KEY_END)

        rE(self.evExpand, curses.KEY_RIGHT, 'l')
        rE(self.evEnter, '\n')
        rE(self.evUnexpand, curses.KEY_LEFT, 'h', 263)

    def setCurY(self, new):
        if new < self.heightreduce:
            new = self.heightreduce
        elif new > self.height:
            new = self.height
        self.__cury = new
    cury = property(lambda self: self.__cury, setCurY)

    def refresh(self):
        if not self.active:
            return

        self.repairIndices()
        self.window.clear()

        logging.debug("TreeWindow: refreshing tree")
        logging.debug("  offset: %s"%(self.offset))
        logging.debug("  height: %s"%(self.height))
        logging.debug("  len(items): %s"%(len(self.items)))
        if len(self.items) == 0:
            self.addLine(self.heightreduce, "No query available")
            for y in xrange(self.heightreduce + 1, self.height):
                try:
                    self.addLine(y)
                except curses.error, e:
                    logging.debug("Catched curses.error on y=%i: %s" % (y, e))
                    break
            return

        curitem = self.items[self.offset]
        blankline = False
        for y in xrange(self.heightreduce, self.height):
            # logging.debug("trying to display '%s'"%(curitem))
            try:
                if not blankline:
                    self.displayItem(curitem.index, y)
                    if curitem.index == self.cursel:
                        self.cury = y
                    curitem = curitem.next
                else:
                    self.addLine(y)
            # TreeErrors are expected and not evil at all.
            except TreeError, e:
                blankline = True
                logging.debug("Catched TreeError: %s"%(e))
            except SizeError, e:
                blankline = True
                logging.error("Catched SizeError: %s"%(e))
            # curses errors are common on the last line
            except curses.error, e:
                logging.error("Catched curses.error: %s"%(e))
                break
            # except:
                # logging.error("Catched unknown error")
                # break

        # we have to call MultiWindow.refresh directly, because
        # ListWindow.refresh would redraw some lines
        MultiWindow.refresh(self)

    # event handlers
    @needsitems
    def evCursorDown(self, event):
        try:
            next = self.items[self.cursel].next
        except TreeError, e:
            self.message("could not get next item")
            return
        self.cursel = next.index

        # I'm at the bottom so I have to recalc self.offset
        if self.cury >= self.height - 2:
            tmp = self.items[self.cursel]
            try:
                for i in xrange(self.height-2):
                    tmp = tmp.prev
                self.offset = tmp.index
            except TreeError:
                self.offset = 0
        else:
            self.cury += 1
        self.refresh()

    @needsitems
    def evCursorUp(self, event):
        try:
            prev = self.items[self.cursel].prev
        except TreeError, e:
            self.message("could not get prev item")
            return
        self.cursel = prev.index

        if self.cury == 1:
            # I'm at the top so I have to recalc self.offset
            self.offset = self.cursel
        else:
            self.cury -=1
        self.refresh()

    @needsitems
    def evPosTop(self, event):
        if self.cursel == 0:
            self.message("reached top of list")
            return
        self.cursel = 0
        self.refresh()
        self.message("cursor home")

    @needsitems
    def evPosEnd(self, event):
        if self.cursel == len(self.items) - 1:
            self.message("reached end of list")
            return

        curitem = self.items[0]
        # run to last artist
        for i in xrange(len(self.items)):
            try:
                curitem = curitem.getNext(childok=False)
            except TreeError:
                logging.debug("evPosEnd: reached end with item %s"%(curitem))
                break
        # find the last expanded entry
        for i in xrange(len(self.items)):
            try:
                curitem = curitem.getNext(childok=True)
            except TreeError:
                logging.debug("evPosEnd: reached end with item %s"%(curitem))
                break

        self.cursel = curitem.index
        # recalc self.offset
        for i in xrange(self.height-2):
            try:
                curitem = curitem.prev
            except TreeError:
                logging.debug("evPosEnd: no prev element %s"%(curitem))
                break

        self.offset = curitem.index
        self.refresh()
        self.message("cursor end")

    def evExpand(self, event):
        self.items[self.cursel].expanded = True
        self.refresh()

    def evUnexpand(self, event):
        if self.items[self.cursel].expanded:
            self.items[self.cursel].expanded = False
        else:
            try:
                self.cursel = self.items[self.cursel].parent.index
            except TreeError, e:
                return
        self.refresh()

    def evRefresh(self, event):
        self.message("refreshed display")

        self.window.clear()
        self.refresh()

    def evEnter(self, event):
        self.items[self.cursel].toggleExpanded()
        self.refresh()
