# vim:fileencoding=utf-8
"""kuechenwindows.py

"""

import threading
import logging
import curses
import time
import ConfigParser
import os.path

from bigstring import WindowedMLS
from eventframework import EventRecipient
from helpers import ms2string, treatURL
from basewindows import SizeError, MultiWindow

class ScrollingText(threading.Thread):
    def __init__(self, name, win, mainapp, intervall=.3, separator=" + "):
        """
        @param win: the curses window
        @param stopflag: a threading.Event object. If set the scrolling loop is
            finished.
        """
        super(ScrollingText, self).__init__(name="ScrollingText")
        self.name = name
        self.__win = win
        self.__inter = intervall
        self.__mainapp = mainapp

        self.__quitflag = mainapp.quitflag
        # a flag indicating that the run loop should pause
        self.__playing = threading.Event()

        self.__sep = separator
        self.__wmls = None

        try:
            self.attr = self.__mainapp.getAttrOption(self.name, "attr",
                    fallback = False)
        except ConfigParser.Error:
            self.attr = 0
        self.attr |= self.__mainapp.getColorAttr(self.name, "color")

    def pause(self):
        self.__playing.clear()

    def play(self):
        self.__playing.set()

    def isPlaying(self):
        return self.__playing.isSet()

    def setText(self, t):
        was_playing = self.__playing.isSet()
        # pause display while changing title
        self.__playing.clear()

        if self.__wmls is None:
            oldind = 0
        else:
            oldind = self.__wmls.index

        self.__wmls = WindowedMLS(t + self.__sep, index=oldind)
        if self.__wmls.height > self.__win.getmaxyx()[0]:
            raise SizeError("New message too high (%i) for window"%(self.__wmls.height))
        if was_playing:
            self.__playing.set()

    def updateText(self, infodict):
        try:
            if infodict.has_key("title"):
                l = "%s - %s" % (infodict.get("artist", "artist"),
                                 infodict.get("title", "title"))
                self.setText(l)
            elif infodict.has_key("url"):
                self.setText(os.path.basename(infodict["url"]))
        except AttributeError, e:
            self.setText("%s" % (e))

    def updateWindow(self):
        width = self.__win.getmaxyx()[1]
        for y, l in enumerate(self.__wmls.getWindow(width)):
            try:
                self.__win.addstr(y, 0, l, self.attr)
            except curses.error:
                pass
        self.__win.refresh()

    def run(self):
        self.updateWindow()
        while not self.__quitflag.isSet():
            self.__playing.wait()
            self.updateWindow()
            self.__wmls.step()
            time.sleep(self.__inter)


class StatusLine(object):
    lrsep = "  "
    def __init__(self, name, window, mainapp):
        self.__win = window
        self.name = name
        self.__mainapp = mainapp

        # texts: right overrule left texts!
        self.__left = []
        self.__right = []
        for i in xrange(self.height):
            self.__left.append('')
            self.__right.append('')

        self.encoding = self.__mainapp.getOption("general", "encoding")

        try:
            self.attr = self.__mainapp.getAttrOption(self.name, "attr")
        except ConfigParser.Error:
            self.attr = 0

        self.attr |= self.__mainapp.getColorAttr(self.name)

    def setWindow(self, newwin):
        self.__win = newwin

    height = property(lambda self: self.__win.getmaxyx()[0])
    width = property(lambda self: self.__win.getmaxyx()[1])

    def setMessage(self, m, y=0, x="l"):
        # logging.debug("StatusLine: setMessage: m='%s', y=%i, x='%s'"%(
                        # m[:25], y, x))
        if y >= self.height:
            raise SizeError("y(%i) index greater than window"
                            " height(%i)!"%(y, self.height))


        if type(m) == unicode:
            m = m.encode(self.encoding, 'replace')
        if x.lower() in ("r", "right"):
            self.__right[y] = m[:self.width]
        else:
            self.__left[y] = m[:self.width]
        self.refresh([y])

    def refresh(self, ylist=None):
        if ylist is None:
            ylist = xrange(self.height)

        for y in ylist:
            l = self.__left[y]
            r = self.lrsep + self.__right[y]
            # r = self.__right[y]
            diff = self.width - len(l) - len(r)
            if diff >= 0:
                line = l + " "*diff + r
            else:
                line = l[:diff-3] + "..." + r
            try:
                self.__win.addstr(y, 0, line[:self.width], self.attr)
                # logging.debug("StatusLine.update: no curses error: y=%i,\n"
                              # "  line='%s'"%(y, line))
            except curses.error:
                # logging.debug("StatusLine.update: catched curses error:"
                              # " y=%i,\n"
                              # "  line='%s'"%(y, line))
                pass
        self.__win.refresh()


class StatusInfoWindow(MultiWindow):
    option_key = "statusinfo"
    color_pair = 7
    def __init__(self, *multiwinargs):
        super(StatusInfoWindow, self).__init__(*multiwinargs)
        # store the medialib ids of current, previous and next 2 tracks
        self.__current = None
        self.__prev = None
        self.__next1 = None
        self.__next2 = None

        self.registerEvent(self.onRefresh, "r")
        self.registerEvent(self.onRatingGood, "g")
        self.registerEvent(self.onRatingBad, "b")

    def onRatingBad(self, ev):
        if self.__current is None:
            logging.error("Cannot rate: no information about current track!")
            return

        id_ = self.__current["id"]
        oldrating = self.__current.get("rating", 0)
        if oldrating <= 0:
            logging.info("Minimal rating value 0 already set!")
            self.message("Minimal rating value reached!")
            return

        self.xmms.medialib_property_set(id_, "rating", oldrating - 1,
                "client/generic")
        self.message("Decreased rating")
        self.refetchData()
        self.refresh()

    def onRatingGood(self, ev):
        if self.__current is None:
            logging.error("Cannot rate: no information about current track!")
            return

        id_ = self.__current["id"]
        oldrating = self.__current.get("rating", 0)
        if oldrating >= 5:
            logging.info("Maximal rating value 5 already set!")
            self.message("Maximal rating value reached!")
            return

        self.xmms.medialib_property_set(id_, "rating", oldrating + 1,
                "client/generic")
        self.message("Increased rating")
        self.refetchData()
        self.refresh()

    def onRefresh(self, ev):
        self.refetchData()
        self.refresh()

    def defaultPlaylistCallback(self, result):
        if not self.active:
            return
        result.wait()
        d = result.value()
        if d["name"] == u"Default":
            self.activate()
            self.refresh()

    cbPlaylistCurrentPos = defaultPlaylistCallback
    cbPlaylistChanged = defaultPlaylistCallback

    def cbCurrentID(self, result):
        if not self.active:
            return
        result.wait()
        self.activate()
        self.refresh()

    def activate(self):
        super(StatusInfoWindow, self).activate()
        self.refetchData()

    def refetchData(self):
        self.__current = None
        r = self.xmms.playback_current_id()
        r.wait()
        r = self.xmms.medialib_get_info(r.value())
        r.wait()
        self.__current = r.value()

        self.__prev = None
        self.__next1 = None
        self.__next2 = None

        r = self.xmms.playlist_list_entries()
        r.wait()
        playlist = r.value()
        # logging.debug("StatusInfoWindow playlist: %s" % playlist)
        r = self.xmms.playlist_current_pos()
        r.wait()
        pos = r.value()
        # logging.debug("StatusInfoWindow pos: %s" % pos)
        if pos is not None:
            curpos = pos["position"]
            if curpos > 0:
                r = self.xmms.medialib_get_info(playlist[curpos - 1])
                r.wait()
                self.__prev = r.value()
            if curpos < len(playlist) - 1:
                r = self.xmms.medialib_get_info(playlist[curpos + 1])
                r.wait()
                self.__next1 = r.value()
            if curpos < len(playlist) - 2:
                r = self.xmms.medialib_get_info(playlist[curpos + 2])
                r.wait()
                self.__next2 = r.value()

    def strhelper(self, info):
        """Return a 'title - artist  duration' string for the given info.
        The string is properly encoded and has the correct length.
        """
        if info.has_key("title"):
            text = u"%s - %s" % (info.get("artist", "Unknown"),
                             info.get("title", "Unknown"))
        else:
            text = info.get("url")
            if text.startswith(u"file"):
                text = os.path.basename(text)
        text = text.encode(self.encoding, "replace")
        if len(text) < self.width:
            text = text.ljust(self.width)
        elif len(text) > self.width:
            text = text[:self.width]
        time = ms2string(info.get("duration", 0))
        return (text[:-len(time)] + time)

    def refresh(self):
        if not self.active:
            return
        self.window.clear()
        # shortcuts
        width = self.width
        emphasized = self.colorattr | curses.A_BOLD

        if self.__current is None:
            try:
                self.addLine(0, "Error: Could not get information from media lib!")
                for y in xrange(1, self.height):
                    self.addLine(y)
            except curses.error, e:
                logging.error("Failed during addstr: %s"  % (e))
            super(StatusInfoWindow, self).refresh()
            return

        try:
            self.addLine(0, "last track:") #, self.colorattr)
            if self.__prev is not None:
                self.addLine(1, (self.strhelper(self.__prev), emphasized))
            else:
                self.addLine(1, "none")
            self.addLine(2)

            self.addLine(3, "now playing:") #, self.colorattr)
            for yoff, key in enumerate(("artist", "album", "title")):
                self.addLine(4 + yoff,
                        key + ": ",
                        (self.__current.get(key, u"unknown"), emphasized))

            url = treatURL(self.__current["url"])

            if url.startswith(u"file://"):
                self.addLine(7, "path: ",
                        (url[7:], emphasized, "left"))
            else:
                self.addLine(7, "url: ",
                        (url, emphasized, "left"))

            self.addLine(8,
                    "duration: ",
                    (ms2string(self.__current.get("duration", 0)), emphasized),
                    "   id: ",
                    ("%d" % (self.__current["id"]), emphasized),
                    "   rating: ",
                    ("%d" % (self.__current.get("rating", 0)), emphasized))
            self.addLine(9)

            self.addLine(10, "next tracks:")
            if self.__next1 is not None:
                self.addLine(11, (self.strhelper(self.__next1), emphasized))
                if self.__next2 is not None:
                    self.addLine(12, (self.strhelper(self.__next2), emphasized))
            else:
                self.addLine(11, "none")
                self.addLine(12)
        except curses.error, e:
            logging.error("Failed during addstr: %s"  % (e))

        for y in xrange(13, self.height):
            try:
                self.addLine(y)
            except curses.error, e:
                logging.error("Failed during addLine: %s"  % (e))
                break

        super(StatusInfoWindow, self).refresh()


class TrackInfoWindow(MultiWindow):
    help = "any key exits"
    def __init__(self, name, parent, distrib, window, mainapp, info):
        super(TrackInfoWindow, self).__init__(name, parent, distrib, window, mainapp)
        self.__info = info

    def handleEvent(self, event):
        self.window.clear()
        super(TrackInfoWindow, self).refresh()
        self.deactivate()
        self.parent.refresh()

    def refresh(self):
        if not self.active:
            return
        width = self.width
        emphasized = self.colorattr | curses.A_BOLD

        def myaddstr(y, prefix, text):
            logging.debug("TrackInfoWindow adding '%s'"%(prefix))
            logging.debug(text)

            self.addStr(y, 0, prefix, self.colorattr)
            if type(text) == unicode:
                # replace unicode BOMs, sometimes they are in the strings,
                # e.g. in comments
                # text = text.replace(u'\ufeff', u'')
                # text = text.replace(u'\uffef', u'')
                # logging.debug(text)
                text = text.encode(self.encoding, 'replace')
            elif type(text) != str:
                text = str(text)

            lines = text.split('\n')
            line = lines[0]
            first = True
            loffs = 0
            for line in lines:
                while line:
                    if first:
                        self.addLine(y, prefix, (line, emphasized, "none"))
                        line = line[width-len(prefix):]
                        first = False
                    else:
                        self.addLine(y + 1 + loffs, (line, emphasized, "none"))
                        line = line[width:]
                        loffs += 1

            return loffs

        def prettyurlhelper(url):
            if url.startswith("file://"):
                return ("Path:     ", treatURL(url)[7:])
            else:
                return ("URL:      ", treatURL(url))

        offset = 0
        for y,(p, t) in enumerate([
                 ("Artist:   ", self.__info.get('artist', 'unknown'))
                ,("Album:    ", self.__info.get('album', 'unknown'))
                ,("Title:    ", self.__info.get('title', 'unknown'))
                ,("Tracknr:  ", self.__info.get('tracknr', 'unknown'))
                ,("Genre:    ", self.__info.get('genre', 'unknown'))
                ,("Comment:  ", self.__info.get('comment', 'unknown'))
                ,("Rating:   ", self.__info.get('rating', 0))
                ,("ID:       ", self.__info.get('id'))
                ,prettyurlhelper(self.__info.get("url", u"unknown"))
                ,("raw URL:  ", self.__info.get('url', 'unknown'))
                ,("Duration: ", ms2string(self.__info.get('duration')))
                ]):
            try:
                offset += myaddstr(y + offset, p, t)
            except curses.error:
                break
        lasty = y + offset
        for y in xrange(lasty + 1, self.height):
            try:
                self.addLine(y)
            except curses.error:
                break
        super(TrackInfoWindow, self).refresh()


class DummyMW(MultiWindow):
    __last = None
    def handleEvent(self, event):
        l = "Catched curses events '%s' '%s'"%(event, self.__last)
        self.__last = event
        self.message(l)
        self.window.addstr(2, 0, l)
        self.refresh()
        self.parent.handleEvent(event)
