#!/usr/bin/env python
# vim:fileencoding=utf-8
"""mainapp.py

"""

import os.path
import threading
import time
import logging
import ConfigParser
import curses
import xmmsclient

from eventframework import EventDistributor, EventRecipient
from bigstring import FigletFrontend
from kuechenwindows import ScrollingText, StatusLine
from helpers import convAttr, convColor, ms2string
from loggingwindow import MyHandler, LoggingWindow

configtemplate="""## default configuration for kuechenstation
[general]
## the file to log to
# logfile = /dev/null

## the encoding used for curses output
## see http://kuechenstation.sourceforge.net/README.html for more information
encoding = %(encoding)s

## the default colors (foreground background)
## can be overwritten (see below)
colors = white black

## the default attribute to indicate the cursor in list or tree windows
cursor attr = reverse

[scrollingtitle]
## the path to the figlet binary
figlet path = %(figlet)s
## the figlet font used
## use figlist(6) to list available fonts
figlet font = standard
attr = reverse
# colors = red blue

[statusline]
attr = reverse

## multi windows

[log]
## the key to switch to this multi window
## can be a string starting with 'KEY_' to define a curses key constant or a
## single character
access key = KEY_F10

[statusinfo]
## use this to define the multi window that is shown on startup
first = true
access key = KEY_F1
# colors = red blue

[playlist]
access key = KEY_F2
## playlist colors and attributes
# colors = yellow blue
# cursor attr = underline
playing attr = bold

[trackinfo]
# colors = black white

[medialib]
access key = KEY_F3
# cursor attr = blink

[streams]
access key = KEY_F4

#[podcasts]
# access key = KEY_F5
## the urls of the podcasts to listen to, separated by commas (',')
# channels = http://www.ndr.de/podcastlink/satirischer_wochenrueckblick.xml,http://www.ndr.de/podcastlink/ndrinfo_streitkraefteundstrategien.xml,http://www.dradio.de/rss/podcast/sendungen/hintergrund/,http://www.dradio.de/rss/podcast/sendungen/kalenderblatt/

#[playlistslist]
# access key = KEY_F6
"""

class MainApp(EventRecipient, EventDistributor):
    sleepint = 0.05
    statusheight = 2
    def __init__(self, screen):
        EventRecipient.__init__(self, None, self)
        EventDistributor.__init__(self, self)

        curses.noecho()
        curses.curs_set(0)
        curses.use_default_colors()
        self.__screen = screen

        # set this if we want to quit
        self.quitflag = threading.Event()
        # a list of XMMS info dictionaries for copying between multi
        # windows.
        self.copybuffer = list()
        self.__playtime = 0.0

        self.xmms_currentid = None
        self.xmms = xmmsclient.XMMS("kuechenstation")

        try:
            self.__init_config()
            self.__init_windows()
            self.__init_xmms()
            self.__init_register_events()
        except ConfigParser.Error, conferr:
            conferr.message += ("\nPerhaps you are using an outdated"
                                " configuration version?")
            raise conferr

    def __init_register_events(self):
        rE = self.registerEvent
        rE(self.evQuit, 'q', 'Q')
        rE(self.evPlayPause, 'p', ' ')
        rE(self.evPlayPause, 'p', ' ', '.')
        rE(self.evVolumeUp, '+', '*')
        rE(self.evVolumeDown, '-', '/')
        rE(self.evSelectMultiwindow,
                *self.__multiwindows.keys())
        rE(self.evSeekForward, curses.KEY_RIGHT)
        rE(self.evSeekBackward, curses.KEY_LEFT)
        rE(self.evSeekForwardLong, curses.KEY_SRIGHT)
        rE(self.evSeekBackwardLong, curses.KEY_SLEFT)
        rE(self.evNextSong, 'n', '>')
        rE(self.evPrevSong, 'b', '<')

    def __init_xmms(self):
        x = self.xmms
        try:
            xmmspath = self.getOption("general", "xmms path")
        except ConfigParser.NoOptionError:
            xmmspath = os.getenv("XMMS_PATH")

        x.connect(xmmspath, self.evQuit)

        # register call backs
        x.broadcast_playback_status(self.cbPlaybackStatus)
        x.broadcast_playback_current_id(self.cbCurrentID)
        x.broadcast_playlist_changed(self.cbPlaylistChanged)
        x.broadcast_playlist_current_pos(self.cbPlaylistCurrentPos)
        x.signal_playback_playtime(self.cbPlaybackPlaytime)

        # init some stuff
        r = x.playback_status()
        r.wait()
        self.cbPlaybackStatus(r)

        r = x.playback_current_id()
        r.wait()
        self.cbCurrentID(r)

    def __init_windows(self):
        H, W = self.__screen.getmaxyx()

        # determine the height of the scrolling title window
        FigletFrontend.figpath = self.getOption("scrollingtitle",
                "figlet path")
        FigletFrontend.figfont = self.getOption("scrollingtitle",
                "figlet font")
        h = FigletFrontend("a").height

        # sub window for the scrolling title window
        subwin = self.__screen.subwin(h, W, 0, 0)
        self.__title = ScrollingText("scrollingtitle", subwin, self,
                intervall=0.3)
        self.__title.setText("just started")

        # sub window for the multi windows
        subwin = self.__screen.subwin(H - h -self.statusheight, W, h, 0)

        # maps access keys to multiwindow instances
        self.__firstmw = None
        self.__currentmw = None
        self.__multiwindows = {}

        # a helper dict to build the default help string
        helpdict = {}

        multiwindows = map(self.mapSection2Class, self.__config.sections())
        multiwindows = filter(lambda x: x is not None, multiwindows)
        for name, cls in multiwindows:
            accesskey = self.getOption(name, "access key")
            if accesskey.startswith("KEY_"):
                helpvals = (name, accesskey[4:])
                accesskey = eval("curses." + accesskey)
            elif len(accesskey) == 1:
                helpvals = (name, accesskey)
                self.__help += "%s: %s  " % (name, accesskey)
            else:
                raise ValueError("option %s.accsess key must be either"
                        " starting with 'KEY_' or a length-1 string!" % (name))
            helpdict[accesskey] = helpvals
            if self.__multiwindows.has_key(accesskey):
                raise ValueError("Multiwindow access key 'curses.%s' assigned"
                        " twice!" % (self.getOption(name,
                                     "access key")))
            try:
                if self.getOption(name, "first").lower() in ("yes", "true"):
                    self.__firstmw = accesskey
            except:
                pass

            mwin = cls(name, self, self, subwin, self)
            self.__multiwindows[accesskey] = mwin

            # set target if we add a logging window
            if cls == LoggingWindow:
                self.__myloghandler.setTargetWindow(mwin)

        logging.info("Available multi windows: %s" % (self.__multiwindows))
        if self.__firstmw is None:
            self.__firstmw = self.__multiwindows.keys()[0]

        self.__help = ""
        for accesskey in sorted(helpdict.iterkeys()):
            name, key = helpdict[accesskey]
            self.__help += "%s: %s  " % (key, name)
        self.__help = self.__help[:-2]

        # sub window for the status line
        subwin = self.__screen.subwin(self.statusheight, W,
                                      H - self.statusheight, 0)
        self.__status = StatusLine("statusline", subwin, self)

    def run(self):
        self.__screen.nodelay(1)
        self.__title.start()
        self.evSelectMultiwindow(self.__firstmw)
        while not self.quitflag.isSet():
            try:
                inp = self.__screen.getch()
                if inp == curses.KEY_RESIZE:
                    self.onResize()
                elif inp != curses.ERR:
                    self.distributeEvent(inp)
                if self.xmms.want_ioout():
                    self.xmms.ioout()
                self.xmms.ioin()
                time.sleep(self.sleepint)
            except KeyboardInterrupt, e:
                logging.error("Failed with error: %s" % (e))
                self.evQuit()

    def onResize(self):
        H, W = self.__screen.getmaxyx()
        logging.debug("onResize: H x W = %i x %i"%(H, W))
        h = FigletFrontend('a').height

        subwin = self.__screen.subwin(H - h - self.statusheight, W,
                                      h, 0)
        logging.debug("multi size: %s"%(subwin.getmaxyx(),))
        for mw in self.__multiwindows.itervalues():
            mw.window = subwin

        subwin = self.__screen.subwin(self.statusheight, W,
                                      H - self.statusheight, 0)
        logging.debug("status size: %s"%(subwin.getmaxyx(),))
        self.__status.setWindow(subwin)
        self.__status.refresh()
        self.__currentmw.refresh()

    def occupy(self, new_recip):
        super(MainApp, self).occupy(new_recip)
        try:
            if new_recip.help is not None:
                self.setHelpText(new_recip.help)
        except AttributeError:
            self.setHelpText(self.__help)

    # {
    # configuration stuff
    #
    @staticmethod
    def mapSection2Class(section):
        """map multi window config section names to MultiWindow classes"""
        cls = None

        if section == "playlist":
            from playlistwindow import PlaylistWindow
            cls = PlaylistWindow
        elif section == "statusinfo":
            from kuechenwindows import StatusInfoWindow
            cls = StatusInfoWindow
        elif section == "medialib":
            from mlibwindow import MedialibQueryWindow
            cls = MedialibQueryWindow
        elif section == "streams":
            from speciallists import StreamListWindow
            cls = StreamListWindow
        elif section == "podcasts":
            from speciallists import PodcastTreeWindow
            cls = PodcastTreeWindow
        elif section == "playlistslist":
            from speciallists import PlaylistListWindow
            cls = PlaylistListWindow
        # elif section == "collections":
            # from speciallists import CollectionTree
            # cls = CollectionTree
        elif section == "log":
            # from loggingwindow import LoggingWindow
            cls = LoggingWindow

        if cls is not None:
            return (section, cls)
        else:
            return None

    @classmethod
    def default_config(cls):
        import locale, subprocess
        encoding = locale.getpreferredencoding()
        po = subprocess.Popen(["which", "figlet"], stdout=subprocess.PIPE)
        if po.wait() == 0:
            figpath = po.stdout.read().strip()
        else:
            raise IOError("Could not find figlet on your system!")
        return configtemplate % ({"encoding":encoding, "figlet" : figpath})

    def __init_config(self):
        cpath = os.path.join(os.path.join(xmmsclient.userconfdir_get(),
                                          'clients'),
                                          'kuechenstation.conf')
        # write a default config file if none exists
        if not os.path.exists(cpath):
            ctext = self.default_config()
            f = open(cpath, "w")
            f.write(ctext)
            f.close()

        self.__config = ConfigParser.ConfigParser()
        self.__config.read(cpath)

        self.__usedcolorpairs = {}

        rootlogger = logging.getLogger()
        rootlogger.setLevel(logging.DEBUG)
        # don't log initial stuff to anything
        rootlogger.addHandler(logging.FileHandler("/dev/null"))

        # init logging
        logfile = self.getOption("general", "logfile", False)
        if logfile is not False:
            rootlogger.addHandler(logging.FileHandler(logfile))

        self.__myloghandler = MyHandler(10)
        logging.getLogger().addHandler(self.__myloghandler)
        logging.debug("logging to '%s'"%(logfile))

    def getOption(self, section, key, default=None):
        logging.debug("getOption(%s, %s, %s)" % (section, key, default))
        if default is not None:
            try:
                return self.__config.get(section, key)
            except ConfigParser.Error, e:
                logging.error("Could not get option %s.%s using default"
                        " '%s': %s" % (section, key, default, e))
                return default
        else:
            return self.__config.get(section, key)

    def getAttrOption(self, multiwindow, key, fallback = True):
        logging.debug("getAttrOption(%s, %s, %s)" % (multiwindow, key,
            fallback))
        attr = None
        try:
            attr = convAttr(self.getOption(multiwindow, key))
        except ConfigParser.Error, e:
            if fallback:
                try:
                    logging.info("No attr '%s' for multi window '%s' defined."
                            " Trying to use default as fallback: %s" % (
                            key, multiwindow, e))
                    attr = convAttr(self.getOption("general", key))

                except ConfigParser.NoOptionError, e:
                    attr = 0
            else:
                logging.info("No attr '%s' for multi window '%s' defined:"
                        " %s" % (key, multiwindow, e))
                raise e

        except Exception, e:
            logging.error("Unexpected error while getting attr '%s' for"
                    " multi window '%s': %s" % (key, multiwindow, e))
            raise e
        return attr

    def getColorAttr(self, window, fallback=True):
        """Returns the curses colorpair attribute."""
        cpair = None
        try:
            cpair = convColor(self.__config.get(window, "colors"))
        except ConfigParser.Error, e:
            if fallback:
                logging.debug("No colorpair for multi window '%s' defined."
                        " Using default colors" % (window))
                cpair = convColor(self.__config.get("general", "colors"))
            else:
                logging.error("No color pair for window %s found!" % (window))
                raise e
        except Exception, e:
            logging.error("Unexpected error while getting color pair for"
                    " multi window '%s': %s" % (window, e))
            raise e

        if not self.__usedcolorpairs.has_key(cpair):
        # assign new color number for color pair 
            cpnumber = len(self.__usedcolorpairs) + 1
            self.__usedcolorpairs[cpair] = cpnumber
            curses.init_pair(cpnumber, cpair[0], cpair[1])

        return curses.color_pair(self.__usedcolorpairs[cpair])
    # }

    # {
    # status window layout:
    # line 0
    # ======
    # setHelpText()                setExtraText()
    # help                         used by multi windows for extra
    #                              information, e.g. the current selected
    #                              playlist position: 6/20 or something
    #
    # line 1
    # ======
    # setStatusText()              setVolatileText()
    # xmms/program status          volatile information:
    #  (e.g. playing, quitting)     * volume change
    #                               * current track time
    #                               * messages from
    #                                 multi windows
    def setStatusText(self, text):
        self.__status.setMessage(text, 1, 'l')

    def setHelpText(self, text):
        logging.debug("setting help text: %s" % (text))
        self.__status.setMessage(text, 0, 'l')

    def setVolatileText(self, text):
        self.__status.setMessage(text, 1, 'r')

    # unused at the moment
    def setExtraText(self, text):
        self.__status.setMessage(text, 0, 'r')

    def updateStatusText(self):
        self.setStatusText("%s (%s)"%(self.__playstatus,
                                      ms2string(self.__playtime)))
    # }

    # {
    # xmms callbacks
    # def cbExample(self, result):
    #     ...
    def cbPlaybackStatus(self, broadcast):
        if broadcast.value() == xmmsclient.PLAYBACK_STATUS_PLAY:
            self.__title.play()
            self.__playstatus = "playing"
            # self.setStatusText("playing")
        elif broadcast.value() == xmmsclient.PLAYBACK_STATUS_PAUSE:
            self.__title.pause()
            self.__playstatus = "paused"
        elif broadcast.value() == xmmsclient.PLAYBACK_STATUS_STOP:
            self.__title.pause()
            self.__playstatus = "stopped"
            self.__playtime = 0.0
        self.updateStatusText()

        for mw in self.__multiwindows.values():
            try:
                mw.cbPlaybackStatus(broadcast)
            except AttributeError:
                pass

    def cbMedialibEntryChanged(self, medialib_infodict):
        infodict = medialib_infodict.value()
        if infodict["id"] == self.xmms_currentid:
            self.__title.updateText(infodict)
        for mw in self.__multiwindows.values():
            try:
                mw.cbMedialibEntryChanged(broadcast)
            except AttributeError:
                pass

    def cbCurrentID(self, medialibid):
        self.xmms_currentid = medialibid.value()
        info = self.xmms.medialib_get_info(self.xmms_currentid)
        info.wait()
        self.__title.updateText(info.value())

        for mw in self.__multiwindows.values():
            try:
                mw.cbCurrentID(medialibid)
            except AttributeError:
                pass

    def cbPlaylistChanged(self, broadcast):
        for mw in self.__multiwindows.values():
            try:
                mw.cbPlaylistChanged(broadcast)
            except AttributeError:
                pass

    def cbPlaylistCurrentPos(self, broadcast):
        for mw in self.__multiwindows.values():
            try:
                mw.cbPlaylistCurrentPos(broadcast)
            except AttributeError:
                pass

    def cbPlaybackPlaytime(self, signal):
        old = self.__playtime
        self.__playtime = signal.value()
        # logging.debug("playtime %i  old %i"%(self.__playtime, old))
        if self.__playtime - old > 100:
            self.updateStatusText()
        time.sleep(.05)
        signal.restart()
    #}

    # {
    # (key) event handlers
    def evQuit(self, event=None):
        logging.info("quitting ... good buy")
        self.setStatusText("quitting")

        # set title window to play so we can leave the thread loop
        self.__title.play()

        self.quitflag.set()

    def evPlayPause(self, event):
        r = self.xmms.playback_status()
        r.wait()
        if r.value() == xmmsclient.PLAYBACK_STATUS_PLAY:
            r2 = self.xmms.playback_pause()
        else:
            r2 = self.xmms.playback_start()
        r2.wait()

    def evVolumeUp(self, event):
        r = self.xmms.playback_volume_get()
        r.wait()
        d = r.value()
        right = d.get('right') + 5
        if right > 100:
            right = 100
        left = d['left'] + 5
        if left > 100:
            left = 100

        r = self.xmms.playback_volume_set('left', left)
        r.wait()
        r = self.xmms.playback_volume_set('right', right)
        r.wait()
        self.setVolatileText('Vol + %s:%s'%(left, right))

    def evVolumeDown(self, event):
        r = self.xmms.playback_volume_get()
        r.wait()
        d = r.value()
        right = d['right'] - 5
        if right < 0:
            right = 0
        left = d['left'] - 5
        if left < 0:
            left = 0

        r = self.xmms.playback_volume_set('left', left)
        r.wait()
        r = self.xmms.playback_volume_set('right', right)
        r.wait()
        self.setVolatileText('Vol - %s:%s'%(left, right))

    def evSelectMultiwindow(self, event):
        newmw = self.__multiwindows[event]
        if newmw == self.__currentmw:
            return

        # self.occupy(newmw)
        if self.__currentmw is not None:
            self.__currentmw.deactivate(False)
        self.__currentmw = newmw
        self.__currentmw.activate()
        logging.debug("current mw: %s"%(self.__currentmw))
        # try:
            # self.setHelpText(self.__currentmw.help)
        # except AttributeError:
            # self.setHelpText("no help")
        self.__currentmw.refresh()

    def evSeekForward(self, event):
        self.xmms.playback_seek_ms_rel(5000).wait()

    def evSeekBackward(self, event):
        self.xmms.playback_seek_ms_rel(-5000).wait()

    def evSeekForwardLong(self, event):
        self.xmms.playback_seek_ms_rel(60000).wait()

    def evSeekBackwardLong(self, event):
        self.xmms.playback_seek_ms_rel(-60000).wait()

    def evNextSong(self, event):
        self.xmms.playlist_set_next_rel(1).wait()
        self.xmms.playback_tickle().wait()

    def evPrevSong(self, event):
        self.xmms.playlist_set_next_rel(-1).wait()
        self.xmms.playback_tickle().wait()
    # }
