# vim:fileencoding=utf-8
"""mlibwindow.py

"""
#   kuechenstation - a curses client for xmms2
#   
#   Copyright (C) 2008  Jan Eike von Seggern <eikevons@yahoo.de>
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import os.path
import re
import curses
import curses.ascii
from xmmsclient import xmmsapi

from basewindows import MultiWindow, TreeWindow, TreeItem, TreeError, SizeError
from kuechenwindows import TrackInfoWindow
from helpers import ms2string, needsitems


class ArtistItem(TreeItem):
    pass

class AlbumItem(TreeItem):
    indent = '  '

class TrackItem(TreeItem):
    indent = '      '
    def getValue(self, key, default=u"unknown"):
        v = self.data[key]
        if v is None:
            return default
        else:
            return v

    def getString(self, width=70):
        title = self.getValue('title')
        if self.getValue('url').startswith('http://'):
            l = u"%s%s"%(self.indent, self.getValue('url'))
        elif title == u"unknown":
            l = u"%s%s"%(self.indent,
                         os.path.basename(self.getValue('url', 'unknown')))
        else:
            artist = self.getValue('artist')
            tracknr = self.getValue('tracknr' ,0)
            l = u"%s%s - %02i - %s"%(self.indent, artist, tracknr, title)
        l = l.encode(self.encoding, 'replace')

        if len(l) < width:
            l = l.ljust(width)
        elif len(l) > width:
            l = l[:width]

        timestr ="%s"%(ms2string(self.data.get('duration')))
        # logging.debug(l)
        l = l[:-len(timestr)] + timestr
        # logging.debug(l)
        return l


class QueryError(Exception):
    pass


class MedialibQueryWindow(TreeWindow):
    option_key = "medialib"
    color_pair = 5
    heightreduce = 1
    # this are non-alphanumeric characters accepted in query input line
    # = for exact searches
    # : for matches
    # ; for separating search descriptions
    # _ and % for SQL globbing
    accepted_textinput = map(ord, ' -_%:=.;')
    # regular expression matches
    # field: desciption
    re_groups = {"field" : "([a-z]*)", "sep" : "([:=])",
                 "value" : "([^:=;&]*)"}
    # print (r"(?:%(field)s%(sep)s%(value)s)"
           # r"(?: *; *%(field)s%(sep)s%(value)s)*"%(re_groups))

    ilre = re.compile(r"(?:%(field)s%(sep)s%(value)s)"
                      r"(?: *; *%(field)s%(sep)s%(value)s)* *"%(re_groups))
    coll = None
    def __init__(self, *multiwinargs):
        super(MedialibQueryWindow, self).__init__(*multiwinargs)

        self.__inputline = 'id:'
        self.__inputactive = True

        # the current selected line in screen coords
        # used to check if we leave the screen with cursor movement

        self.__init_register_events()

    def __init_register_events(self):
        rE = self.registerEvent

        rE(self.evSwitchFocus, curses.ascii.TAB, ':')
        rE(self.evBackspace, 263)

        rE(self.evInfo, 'i')

        rE(self.evAdd, 'a')
        rE(self.evRemove, 'x')
        rE(self.evYank, "y")

    def __build_artist_dict(self, infolist):
        artists = {}
        # fill temporary artists dictionary
        for info in infolist:
            # use lowercase artists for sorting
            t = info['artist']
            if t is None:
                artkey = 'unknown'
            else:
                artkey = t.lower().replace(' ', '')

            t =info['album']
            if t is None:
                albkey = 'unknown'
            else:
                albkey = t.strip()

            # tartist: dictionary of (album name:track list) pairs,
            # one per artist
            save_tartist = False
            # tracks: list with the tracks, one per album
            save_tracks = False

            if not artists.has_key(artkey):
                tartist = dict()
                save_tartist = True
            else:
                tartist = artists[artkey]

            if not tartist.has_key(albkey):
                tracks = list()
                save_tracks = True
            else:
                tracks = tartist[albkey]

            tracks.append(info)

            if save_tracks:
                tartist[albkey] = tracks
            if save_tartist:
                artists[artkey] = tartist

        return artists

    def loadCollection(self):
        # self.tracks = []
        r = self.xmms.coll_query_infos(self.coll,
                    ['artist', 'album', 'title', 'tracknr', 'id', 'url', 'duration'])
        r.wait()
        infolist = r.value()
        if not infolist:
            raise QueryError("nothing found")

        artists = self.__build_artist_dict(infolist)

        # delete the old search
        self.items = []

        lastart = None
        for artname in sorted(artists.iterkeys()):
            albdict = artists[artname]
            curart = ArtistItem(artname, self.encoding, len(self.items),
                                prev=lastart)
            if lastart:
                lastart.setNext(curart)
            lastart = curart
            self.items.append(curart)

            lastalb = None
            # for albname in sorted(albdict.iterkeys()):
            for albname in albdict.iterkeys():
                curalb = AlbumItem(albname, self.encoding, len(self.items),
                                   parent=curart, prev=lastalb)
                if lastalb:
                    lastalb.setNext(curalb)
                lastalb = curalb
                self.items.append(curalb)

                songlist = albdict[albname]
                songlist.sort(key=lambda d: d.get('tracknr'))
                test = set(map(lambda d: d.get('tracknr'), songlist))
                if len(songlist)*1./len(test) >= 1.8:
                    logging.warning("inserting album '%s': same tracknumbers"
                                    " occured multiple times, using URL as"
                                    " sorting criterium.\nPerhaps you"
                                    " should re-tag your files!"%(albname))
                    songlist.sort(key=lambda d: d.get('url'))

                lasttrack = None
                for trackinfo in songlist:
                    curtrack = TrackItem(trackinfo, self.encoding, len(self.items),
                                          parent=curalb, prev=lasttrack)
                    if lasttrack:
                        lasttrack.setNext(curtrack)
                    lasttrack = curtrack
                    self.items.append(curtrack)

                    curalb.addChildren(curtrack)
                    # self.tracks.append(curtrack)

                curart.addChildren(curalb)

    def setCurY(self, new):
        if new < 1:
            new = 1
        elif new > self.height:
            new = self.height
        self.__cury = new
    cury = property(lambda self: self.__cury, setCurY)

    def handleEvent(self, event):
        if (self.__inputactive and
            (curses.ascii.isalnum(event) or event in self.accepted_textinput)):
            self.__inputline += chr(event)
            logging.debug("input line character '%i'"%(event))
            self.refreshInputline()
            MultiWindow.refresh(self)
            # super(MedialibQueryWindow, self).refresh()
            return

        logging.debug("other character '%i'"%(event))
        super(MedialibQueryWindow, self).handleEvent(event)

    def refresh(self):
        if not self.active:
            return

        super(MedialibQueryWindow, self).refresh()
        self.refreshInputline()
        MultiWindow.refresh(self)

    def refreshInputline(self):
        prompt = ">>"

        # short cut for the color attribute
        ca = self.colorattr

        # attribute for active prompt
        if self.__inputactive:
            inpact = curses.A_REVERSE
        else:
            inpact = 0
        self.addLine(0,
                (prompt, ca|inpact),
                " ",
                (self.__inputline, ca|inpact, "left"))

        # self.window.addstr(0, 0, prompt, ca|inpact)
        # self.window.addstr(0, len(prompt), ' ', ca)
        # xpos = len(prompt) + 1

        # the width remaining for the input line
        # keep at the end one free character for a marked space
        # ilwidth = self.width - xpos - 1
        # if len(self.__inputline) <= ilwidth:
            # self.window.addstr(0, xpos, self.__inputline, ca)
            # xpos += len(self.__inputline)
        # else:
            # self.window.addstr(0, xpos,
                               # '...' + self.__inputline[-ilwidth+3:], ca)
            # xpos += ilwidth
            # # xpos += len('...' +  self.__inputline[-ilwidth+3:])
        # # print a marker for the position of the cursor
        # self.window.addstr(0, xpos, ' ', ca|inpact)
        # xpos += 1
        # if self.width - xpos > 0:
            # self.window.addstr(0, xpos, ' ' * (self.width - xpos))

    @staticmethod
    def completeField(f):
        if f.startswith('al'):
            return "album"
        if f.startswith('a'):
            return "artist"
        if f.startswith('tr'):
            return "tracknr"
        if f.startswith('t'):
            return "title"
        if f.startswith('g'):
            return "genre"
        if f.startswith("u"):
            return "url"
        if f.startswith("i"):
            return "id"
        if f.startswith("l"):
            return "list"
        if f.startswith("c"):
            return "coll"
        return f

    def parseInputline(self):
        m = self.ilre.search(self.__inputline)
        # m = self.ilre.search(line)
        if m is None:
            self.message("error: could not match input line")
            logging.error("parseInputline: could not match"
                          " input line '%s'"%(self.__inputline))
            raise QueryError("query string did not match!")

        g = m.groups()
        coll = None
        logging.debug("re.match groups: %s"%(g,))
        for i in xrange(0, len(g), 3):
            if g[i] is None:
                break
            f = self.completeField(g[i])
            s = g[i+1]
            v = g[i+2]
            if f == "list":
                r = self.xmms.coll_get(v, 'Playlists')
                r.wait()
                curcoll = r.value()
            elif f == "coll":
                r = self.xmms.coll_get(v, 'Collections')
                r.wait()
                curcoll = r.value()
            else:
                if s != '=':
                    v = v.strip()
                    if not v.startswith("%"):
                        v = "%" + v
                    if not v.endswith("%"):
                        v = v + "%"
                logging.debug("adding Match(field=%s, value=%s)"%(f, v))
                curcoll = xmmsapi.Match(field=f, value=v)
            if curcoll is None:
                continue
            #TODO: handle different group separators
            if coll is None:
                coll = curcoll
            else:
                coll &= curcoll

        logging.debug("collection: %s"%(coll))
        if coll is not None:
            logging.debug("  operands: %s"%(coll.operands))
        self.coll = coll

    def buildAndLoadCollection(self):
        self.message("querying medialib")
        try:
            self.parseInputline()
            logging.info("built collection: %s"%(self.coll))
            self.loadCollection()
            logging.info("loaded collection: %s"%(self.items))

        except QueryError, e:
            self.__inputactive = True
            self.message("error: %s"%(e))

        else:
            self.offset = 0
            self.cursel = 0
            self.__inputactive = False
            self.message("list active")
            self.refresh()

    # TODO:
    def searchCollection(self, pattern, namespace):
        pass

    # event handlers
    @needsitems
    def evCursorDown(self, event):
        if self.__inputactive:
            return
        self.message("list active")
        super(MedialibQueryWindow, self).evCursorDown(event)

    @needsitems
    def evCursorUp(self, event):
        if self.__inputactive:
            return
        self.message("list active")
        super(MedialibQueryWindow, self).evCursorUp(event)

    @needsitems
    def evPosTop(self, event):
        if self.__inputactive:
            return
        super(MedialibQueryWindow, self).evPosTop(event)

    @needsitems
    def evPosEnd(self, event):
        if self.cursel == len(self.items) - 1:
            self.message("reached end of list")
            return
        if self.__inputactive:
            return
        super(MedialibQueryWindow, self).evPosEnd(event)

    def evSwitchFocus(self, event):
        logging.info("switching focus")
        # self.__inputactive = not self.__inputactive
        if self.__inputactive:
            self.__inputactive = False
            self.message("list active")
        else:
            self.__inputactive = True
            self.message("input active")
        self.refresh()

    def evEnter(self, event):
        """
        not in input mode: toggle expanded property of current element
        in input mode: start a query and enter list mode
        """
        if self.__inputactive:
            # self.buildAndLoadCollection()
            self.message("querying medialib")
            try:
                self.parseInputline()
                logging.info("built collection: %s"%(self.coll))
                self.loadCollection()
                logging.info("loaded collection: %s"%(self.items))

            except QueryError, e:
                self.__inputactive = True
                self.message("error: %s"%(e))

            else:
                self.offset = 0
                self.cursel = 0
                self.__inputactive = False
                self.message("list active")
                self.refresh()
        else:
            # TreeWindow.evEnter(self, event)
            super(MedialibQueryWindow, self).evEnter(event)

    def evExpand(self, event):
        if self.__inputactive:
            return
        self.message("list active")
        TreeWindow.evExpand(self, event)
        super(MedialibQueryWindow, self).evExpand(event)

    def evUnexpand(self, event):
        if self.__inputactive:
            return
        self.message("list active")
        TreeWindow.evUnexpand(self, event)
        # super(MedialibQueryWindow, self).evExpand(event)

    def evBackspace(self, event):
        if self.__inputactive:
            logging.debug("Backspace in input mode")
            logging.debug("old input line: '%s'"%(self.__inputline))
            if len(self.__inputline) > 0:
                self.__inputline = self.__inputline[:-1]
                logging.debug("new input line: '%s'"%(self.__inputline))
                self.refresh()
        else:
            logging.debug("Backspace in list mode")
            self.evUnexpand(event)

    def evRefresh(self, event):
        self.message("refreshed display")

        self.window.clear()
        self.refresh()

    def evClear(self, event):
        if self.__inputactive:
            return
        self.__inputline = ""
        self.__inputactive = True
        self.items = []
        self.refresh()

    def evInfo(self, event):
        curitem = self.items[self.cursel]
        if type(curitem) != TrackItem:
            return

        curid = curitem.data['id']
        r = self.xmms.medialib_get_info(curid)
        r.wait()

        tiw = TrackInfoWindow("trackinfo", self, self.distributor,
                self.window, self.mainapp, r.value())
        tiw.activate()
        tiw.refresh()

    @needsitems
    def evAdd(self, event):
        if self.__inputactive:
            return
        self.message("list active")

        cur = self.items[self.cursel]
        if type(cur) == TrackItem:
            self.xmms.playlist_add_id(cur.data['id']).wait()
            logging.debug("added id %s to playlist"%(cur.data['id']))
            self.message("adding track %s"%(cur.data['title']))
        elif type(cur) == AlbumItem:
            self.message("adding album %s"%(cur.data))
            logging.debug("adding album '%s' to playlist"%(cur.data))
            for track in cur.children:
                self.xmms.playlist_add_id(track.data['id']).wait()
                logging.debug("added id %s to playlist"%(track.data['id']))

    def evRemove(self, event):
        if self.__inputactive:
            return
        self.message("list active")
        cur = self.items[self.cursel]
        if type(cur) == TrackItem:
            id = cur.data['id']
            self.xmms.medialib_remove_entry(id).wait()
            logging.debug("removed id %s from medialib"%(id))
            self.message("removed id %s"%(id))
            self.buildAndLoadCollection()
        elif type(cur) == AlbumItem:
            for track in cur.children:
                self.xmms.medialib_remove_entry(track.data['id']).wait()
                logging.debug("removed id %s"%(track.data['id']))
            self.message("removed album %s"%(cur.data))
            logging.debug("removed album '%s'"%(cur.data))
            self.loadCollection()
            self.refresh()

    @needsitems
    def evYank(self, event):
        if self.__inputactive:
            return

        cur = self.items[self.cursel]
        if type(cur) == TrackItem:
            self.mainapp.copybuffer = [cur.data]
            self.message("yanked id %i" % (cur.data["id"]))
        elif type(cur) == AlbumItem:
            self.mainapp.copybuffer = [track.data for track in cur.children]
            self.message("yanked ids %s"
                    % ([track.data["id"] for track in cur.children]))
