# vim:fileencoding=utf-8
"""speciallists.py

"""

import logging

from xmmsclient import xmmsapi

from basewindows import ListWindow, ListItem, TreeWindow, TreeItem
from kuechenwindows import TrackInfoWindow
from helpers import needsitems

class PlaylistListItem(ListItem):
    def getString(self, width):
        l = self.data.encode(self.encoding, 'replace')

        # do all the size stuff with encoded ascii strings
        # if len(l) < width:
            # l = l.ljust(width)
        if len(l) > width:
            l = '...' + l[-width+3:]
        return l

class PlaylistListWindow(ListWindow):
    def __init__(self, *multiwinargs):
        super(PlaylistListWindow, self).__init__(*multiwinargs)

        self.__init_register_events()

    def __init_register_events(self):
        rE = self.registerEvent

        rE(None, ' ')
        rE(self.evAdd, '\n', 'a')
        # rE(self.evInfo, 'i')

    def activate(self):
        super(PlaylistListWindow, self).activate()
        r = self.xmms.playlist_list()
        r.wait()
        infolist = r.value()
        self.items = []
        for info in infolist:
            if info[0] != "_":
                self.items.append(PlaylistListItem(info, self.encoding))

    @needsitems
    def evAdd(self, event):
        cur = self.items[self.cursel]
        self.xmms.playlist_load(cur.data).wait()


class StreamListItem(ListItem):
    def getString(self, width):
        l = self.data['url'].encode(self.encoding, 'replace')

        # do all the size stuff with encoded ascii strings
        # if len(l) < width:
            # l = l.ljust(width)
        if len(l) > width:
            l = '...' + l[-width+3:]
        return l

class StreamListWindow(ListWindow):
    def __init__(self, *multiwinargs):
        super(StreamListWindow, self).__init__(*multiwinargs)

        self.__init_register_events()

    def __init_register_events(self):
        rE = self.registerEvent

        rE(None, ' ')
        rE(self.evAdd, '\n', 'a')
        rE(self.evInfo, 'i')
        rE(self.evYank, 'y')

    def activate(self):
        super(StreamListWindow, self).activate()
        files = xmmsapi.Match(field="url", value="file%")
        smb = xmmsapi.Match(field="url", value="smb%")
        coll = xmmsapi.Complement(smb | files)
        r = self.xmms.coll_query_infos(coll, ['id', 'url'])
        r.wait()
        infolist = r.value()
        self.items = []
        for info in infolist:
            self.items.append(StreamListItem(info, self.encoding))

    @needsitems
    def evAdd(self, event):
        cur = self.items[self.cursel]
        self.xmms.playlist_add_id(cur.data['id']).wait()

    @needsitems
    def evInfo(self, event):
        curid = self.items[self.cursel].data['id']
        r = self.xmms.medialib_get_info(curid)
        r.wait()

        # tiw = TrackInfoWindow(self, self.distributor, self.window, self.xmms,
                              # r.value())
        tiw = TrackInfoWindow("trackinfo", self, self.distributor,
                self.window, self.mainapp, r.value())
        tiw.activate()
        tiw.refresh()

    @needsitems
    def evYank(self, event):
        self.mainapp.copybuffer = [self.items[self.cursel].data]
        self.message("yanked stream")


class PodcastError(Exception):
    pass

class PodcastChannel(TreeItem):
    pass

class PodcastItem(TreeItem):
    indent = '    '
    def getString(self, width=70):
        l = u"%s%s"%(self.indent, self.data['title'])
        l = l.encode(self.encoding, 'replace')
        if len(l) < width:
            l = l.ljust(width)
        elif len(l) > width:
            l = l[:width]
        return l

    def getUrl(self):
        for encl in self.data['enclosures']:
            if encl['type'].startswith("audio"):
                return encl['href']
        raise PodcastError("No valid enclosure url for title '%s'"%(
                           self.data['title']))
    url = property(getUrl)


class PodcastTreeWindow(TreeWindow):
    def __init__(self, *multiwinargs):
        super(PodcastTreeWindow, self).__init__(*multiwinargs)
        try:
            self.__urls = self.mainapp.getOption(self.name,
                    'channels').split(',')
        except Exception, e:
            logging.error("Error loading podcast channel urls: %s"%(e))
            self.__urls = []
        self.items = []

        self.__init_register_events()

    def __init_register_events(self):
        rE = self.registerEvent

        rE(self.evAdd, 'a')
        rE(self.evRefresh, 'r')

    def reloadChannels(self):
        self.items = []
        lastchannel = None
        self.message("reloading feeds")

        try:
            import feedparser
        except ImportError:
            self.message("failed to import module feedparser")
            logging.error("failed to import module feedparser")
            self.items.append(PodcastChannel(
                                "Failed to import module feedparser.",
                                self.encoding, len(self.items)))
            return

        for url in self.__urls:
            data = feedparser.parse(url)
            try:
                curch = PodcastChannel(data['feed']['title'], self.encoding,
                                       len(self.items), prev=lastchannel)
            except KeyError:
                curch = PodcastChannel("Failed to load %s"%(url), self.encoding,
                                       len(self.items), prev=lastchannel)
                logging.error("failed to load url %s"%(url))
                data["entries"] = []
            if lastchannel:
                lastchannel.next = curch
            lastchannel = curch
            self.items.append(curch)

            lastentry = None
            for entry in data["entries"]:
                curentry = PodcastItem(entry, self.encoding,
                                       len(self.items), parent=curch,
                                       prev=lastentry)
                if lastentry:
                    lastentry.next = curentry
                lastentry = curentry
                curch.addChildren(curentry)
                self.items.append(curentry)

        if len(self.items) == 0:
            self.items.append(PodcastChannel("no channels", self.encoding,
                                             len(self.items)))

    def activate(self):
        super(PodcastTreeWindow, self).activate()
        # if len(self.items) == 0:
            # self.reloadChannels()

    def evAdd(self, event):
        cur = self.items[self.cursel]
        if type(cur) != PodcastItem:
            self.message("only podcast items can be added!")
            logging.info("only podcast items can be added!")
            return
        logging.info("adding url '%s' to playlist"%(cur.url))
        self.xmms.playlist_add_url(cur.url).wait()
        self.message("added %s"%(cur.url))

    def evRefresh(self, event):
        self.reloadChannels()


# TODO build the collection tree
class CategoryNode(TreeItem):
    pass

class CollectionNode(TreeItem):
    indent = '  '


class CollectionTree(TreeWindow):
    pass
